@@ -26,8 +26,8 @@ defprotocol Inspect do
       defimpl Inspect, for: MapSet do
         import Inspect.Algebra
 
-        def inspect(dict, opts) do
-          concat(["#MapSet<", to_doc(MapSet.to_list(dict), opts), ">"])
+        def inspect(map_set, opts) do
+          concat(["#MapSet<", to_doc(MapSet.to_list(map_set), opts), ">"])
         end
       end
 
@@ -157,7 +157,7 @@ defimpl Inspect, for: BitString do
   defp each_bit(bitstring, _counter, opts) do
     size = bit_size(bitstring)
     <<h::size(size)>> = bitstring
-    Inspect.Integer.inspect(h, opts) <> "::size(" <> Integer.to_string(size) <> ")"
+    concat(Inspect.Integer.inspect(h, opts), "::size(" <> Integer.to_string(size) <> ")")
   end
 
   @compile {:inline, decrement: 1}
@@ -454,16 +454,17 @@ defimpl Inspect, for: Any do
   end
 
   def inspect(map, name, opts) do
-    # Use the :limit option and an extra element to force
-    # `container_doc/6` to append "...".
-    opts = %{opts | limit: min(opts.limit, map_size(map))}
-    map = Map.to_list(map) ++ ["..."]
-
+    map = Map.to_list(map) ++ [:...]
     open = color("#" <> name <> "<", :map, opts)
     sep = color(",", :map, opts)
     close = color(">", :map, opts)
 
-    container_doc(open, map, close, opts, &Inspect.List.keyword/2, separator: sep, break: :strict)
+    fun = fn
+      {key, value}, opts -> Inspect.List.keyword({key, value}, opts)
+      :..., _opts -> "..."
+    end
+
+    container_doc(open, map, close, opts, fun, separator: sep, break: :strict)
   end
 end
 
