@@ -10,6 +10,7 @@ defmodule Module.Types.Unify do
   #   {:atom, atom} < :atom
   #   :integer
   #   :float
+  #   :binary
   #   :pid
   #   :port
   #   :reference
@@ -20,65 +21,37 @@ defmodule Module.Types.Unify do
   #   {:tuple, size, [type]} < :tuple
   #   {:union, [type]}
   #   {:map, [{:required | :optional, key_type, value_type}]}
+  #   {:fun, [{params, return}]}
   #
   # Once new types are added, they should be considered in:
   #
   #   * unify (all)
   #   * format_type (all)
   #   * subtype? (subtypes only)
-  #   * has_unbound_var? (composite only)
   #   * recursive_type? (composite only)
+  #   * collect_var_indexes (composite only)
+  #   * lift_types (composite only)
+  #   * flatten_union (composite only)
+  #   * walk (composite only)
   #
 
   @doc """
   Unifies two types and returns the unified type and an updated typing context
   or an error in case of a typing conflict.
   """
-  def unify(source, target, stack, context) do
-    case do_unify(source, target, stack, context) do
-      {:ok, type, context} ->
-        {:ok, type, context}
-
-      {:error, reason} ->
-        if stack.context == :pattern do
-          case do_unify(target, source, stack, context) do
-            {:ok, type, context} ->
-              {:ok, type, context}
-
-            {:error, _} ->
-              {:error, reason}
-          end
-        else
-          {:error, reason}
-        end
-    end
-  end
-
-  defp do_unify(same, same, _stack, context) do
+  def unify(same, same, _stack, context) do
     {:ok, same, context}
   end
 
-  defp do_unify(type, {:var, var}, stack, context) do
-    case context.types do
-      %{^var => {:var, var_type}} ->
-        do_unify(type, {:var, var_type}, stack, context)
-
-      %{} ->
-        unify_var(var, type, stack, context, _var_source = false)
-    end
+  def unify(type, {:var, var}, stack, context) do
+    unify_var(var, type, stack, context, _var_source = false)
   end
 
-  defp do_unify({:var, var}, type, stack, context) do
-    case context.types do
-      %{^var => {:var, var_type}} ->
-        do_unify({:var, var_type}, type, stack, context)
-
-      %{} ->
-        unify_var(var, type, stack, context, _var_source = true)
-    end
+  def unify({:var, var}, type, stack, context) do
+    unify_var(var, type, stack, context, _var_source = true)
   end
 
-  defp do_unify({:tuple, n, sources}, {:tuple, n, targets}, stack, context) do
+  def unify({:tuple, n, sources}, {:tuple, n, targets}, stack, context) do
     result =
       map_reduce_ok(Enum.zip(sources, targets), context, fn {source, target}, context ->
         unify(source, target, stack, context)
@@ -90,33 +63,41 @@ defmodule Module.Types.Unify do
     end
   end
 
-  defp do_unify({:list, source}, {:list, target}, stack, context) do
+  def unify({:list, source}, {:list, target}, stack, context) do
     case unify(source, target, stack, context) do
       {:ok, type, context} -> {:ok, {:list, type}, context}
       {:error, reason} -> {:error, reason}
     end
   end
 
-  defp do_unify({:map, source_pairs}, {:map, target_pairs}, stack, context) do
+  def unify({:map, source_pairs}, {:map, target_pairs}, stack, context) do
     unify_maps(source_pairs, target_pairs, stack, context)
   end
 
-  defp do_unify(source, :dynamic, _stack, context) do
+  def unify(source, :dynamic, _stack, context) do
     {:ok, source, context}
   end
 
-  defp do_unify(:dynamic, target, _stack, context) do
+  def unify(:dynamic, target, _stack, context) do
     {:ok, target, context}
   end
 
-  defp do_unify(source, target, stack, context) do
+  def unify({:union, types}, target, stack, context) do
+    unify_result =
+      map_reduce_ok(types, context, fn type, context ->
+        unify(type, target, stack, context)
+      end)
+
+    case unify_result do
+      {:ok, types, context} -> {:ok, to_union(types, context), context}
+      {:error, context} -> {:error, context}
+    end
+  end
+
+  def unify(source, target, stack, context) do
     cond do
-      # This condition exists to handle unions with unbound vars.
-      # TODO: handle unions properly. Note we can easily unify
-      # "union < type" even if union has vars as the vars must be
-      # type
-      (match?({:union, _}, source) and has_unbound_var?(source, context)) or
-          (match?({:union, _}, target) and has_unbound_var?(target, context)) ->
+      # TODO: This condition exists to handle unions with unbound vars.
+      match?({:union, _}, target) and has_unbound_var?(target, context) ->
         {:ok, source, context}
 
       subtype?(source, target, context) ->
@@ -127,14 +108,14 @@ defmodule Module.Types.Unify do
     end
   end
 
-  defp unify_var(var, :dynamic, _stack, context, _var_source?) do
+  def unify_var(var, :dynamic, _stack, context, _var_source?) do
     {:ok, {:var, var}, context}
   end
 
-  defp unify_var(var, type, stack, context, var_source?) do
+  def unify_var(var, type, stack, context, var_source?) do
     case context.types do
       %{^var => :unbound} ->
-        context = refine_var(var, type, stack, context)
+        context = refine_var!(var, type, stack, context)
         stack = push_unify_stack(var, stack)
 
         if recursive_type?(type, [], context) do
@@ -147,6 +128,31 @@ defmodule Module.Types.Unify do
           {:ok, {:var, var}, context}
         end
 
+      %{^var => {:var, new_var} = var_type} ->
+        unify_result =
+          if var_source? do
+            unify(var_type, type, stack, context)
+          else
+            unify(type, var_type, stack, context)
+          end
+
+        case unify_result do
+          {:ok, type, context} ->
+            {:ok, type, context}
+
+          {:error, {type, reason, %{traces: error_traces} = error_context}} ->
+            old_var_traces = Map.get(context.traces, new_var, [])
+            new_var_traces = Map.get(error_traces, new_var, [])
+            add_var_traces = Enum.drop(new_var_traces, -length(old_var_traces))
+
+            error_traces =
+              error_traces
+              |> Map.update(var, add_var_traces, &(add_var_traces ++ &1))
+              |> Map.put(new_var, old_var_traces)
+
+            {:error, {type, reason, %{error_context | traces: error_traces}}}
+        end
+
       %{^var => var_type} ->
         # Only add trace if the variable wasn't already "expanded"
         context =
@@ -170,7 +176,7 @@ defmodule Module.Types.Unify do
             {:ok, {:var, var}, context}
 
           {:ok, res_type, context} ->
-            context = refine_var(var, res_type, stack, context)
+            context = refine_var!(var, res_type, stack, context)
             {:ok, {:var, var}, context}
 
           {:error, reason} ->
@@ -182,7 +188,7 @@ defmodule Module.Types.Unify do
   # * All required keys on each side need to match to the other side.
   # * All optional keys on each side that do not match must be discarded.
 
-  defp unify_maps(source_pairs, target_pairs, stack, context) do
+  def unify_maps(source_pairs, target_pairs, stack, context) do
     {source_required, source_optional} = split_pairs(source_pairs)
     {target_required, target_optional} = split_pairs(target_pairs)
 
@@ -213,7 +219,7 @@ defmodule Module.Types.Unify do
     end
   end
 
-  defp unify_source_required(source_required, target_pairs, stack, context) do
+  def unify_source_required(source_required, target_pairs, stack, context) do
     map_reduce_ok(source_required, context, fn {source_key, source_value}, context ->
       Enum.find_value(target_pairs, fn {target_kind, target_key, target_value} ->
         with {:ok, key, context} <- unify(source_key, target_key, stack, context) do
@@ -233,7 +239,7 @@ defmodule Module.Types.Unify do
     end)
   end
 
-  defp unify_target_required(target_required, source_pairs, stack, context) do
+  def unify_target_required(target_required, source_pairs, stack, context) do
     map_reduce_ok(target_required, context, fn {target_key, target_value}, context ->
       Enum.find_value(source_pairs, fn {source_kind, source_key, source_value} ->
         with {:ok, key, context} <- unify(source_key, target_key, stack, context) do
@@ -253,7 +259,7 @@ defmodule Module.Types.Unify do
     end)
   end
 
-  defp unify_source_optional(source_optional, target_optional, stack, context) do
+  def unify_source_optional(source_optional, target_optional, stack, context) do
     flat_map_reduce_ok(source_optional, context, fn {source_key, source_value}, context ->
       Enum.find_value(target_optional, fn {target_key, target_value} ->
         with {:ok, key, context} <- unify(source_key, target_key, stack, context) do
@@ -273,7 +279,7 @@ defmodule Module.Types.Unify do
     end)
   end
 
-  defp unify_target_optional(target_optional, source_optional, stack, context) do
+  def unify_target_optional(target_optional, source_optional, stack, context) do
     flat_map_reduce_ok(target_optional, context, fn {target_key, target_value}, context ->
       Enum.find_value(source_optional, fn {source_key, source_value} ->
         with {:ok, key, context} <- unify(source_key, target_key, stack, context) do
@@ -302,7 +308,7 @@ defmodule Module.Types.Unify do
     {required, optional}
   end
 
-  defp error(type, reason, context), do: {:error, {type, reason, context}}
+  def error(type, reason, context), do: {:error, {type, reason, context}}
 
   @doc """
   Push expression to stack.
@@ -406,10 +412,21 @@ defmodule Module.Types.Unify do
     %{stack | unify_stack: [var | stack.unify_stack]}
   end
 
+  @doc """
+  Restores the variable information from the old context into new context.
+  """
+  def restore_var!(var, new_context, old_context) do
+    %{^var => type} = old_context.types
+    %{^var => trace} = old_context.traces
+    types = Map.put(new_context.types, var, type)
+    traces = Map.put(new_context.traces, var, trace)
+    %{new_context | types: types, traces: traces}
+  end
+
   @doc """
   Set the type for a variable and add trace.
   """
-  def refine_var(var, type, stack, context) do
+  def refine_var!(var, type, stack, context) do
     types = Map.put(context.types, var, type)
     context = %{context | types: types}
     trace_var(var, type, stack, context)
@@ -471,36 +488,67 @@ defmodule Module.Types.Unify do
     end)
   end
 
+  defp recursive_type?({:fun, clauses}, parents, context) do
+    Enum.any?(clauses, fn {args, return} ->
+      Enum.any?([return | args], &recursive_type?(&1, [clauses | parents], context))
+    end)
+  end
+
   defp recursive_type?(_other, _parents, _context) do
     false
   end
 
+  @doc """
+  Collects all type vars recursively.
+  """
+  def collect_var_indexes(type, context, acc \\ %{}) do
+    {_type, indexes} =
+      walk(type, acc, fn
+        {:var, var}, acc ->
+          case acc do
+            %{^var => _} ->
+              {{:var, var}, acc}
+
+            %{} ->
+              case context.types do
+                %{^var => :unbound} ->
+                  {{:var, var}, Map.put(acc, var, true)}
+
+                %{^var => type} ->
+                  {{:var, var}, collect_var_indexes(type, context, Map.put(acc, var, true))}
+              end
+          end
+
+        other, acc ->
+          {other, acc}
+      end)
+
+    indexes
+  end
+
   @doc """
   Checks if the type has a type var.
   """
-  def has_unbound_var?({:var, var}, context) do
-    case context.types do
-      %{^var => :unbound} -> true
-      %{^var => type} -> has_unbound_var?(type, context)
-    end
-  end
+  def has_unbound_var?(type, context) do
+    walk(type, :ok, fn
+      {:var, var}, acc ->
+        case context.types do
+          %{^var => :unbound} ->
+            throw(:has_unbound_var?)
 
-  def has_unbound_var?({:tuple, _, args}, context),
-    do: Enum.any?(args, &has_unbound_var?(&1, context))
+          %{^var => type} ->
+            has_unbound_var?(type, context)
+            {{:var, var}, acc}
+        end
 
-  def has_unbound_var?({:union, args}, context),
-    do: Enum.any?(args, &has_unbound_var?(&1, context))
-
-  def has_unbound_var?({:list, arg}, context),
-    do: has_unbound_var?(arg, context)
-
-  def has_unbound_var?({:map, pairs}, context) do
-    Enum.any?(pairs, fn {_, key, value} ->
-      has_unbound_var?(key, context) or has_unbound_var?(value, context)
+      other, acc ->
+        {other, acc}
     end)
-  end
 
-  def has_unbound_var?(_type, _context), do: false
+    false
+  catch
+    :throw, :has_unbound_var? -> true
+  end
 
   @doc """
   Returns true if it is a singleton type.
@@ -556,7 +604,7 @@ defmodule Module.Types.Unify do
   def subtype?({:tuple, n, left_types}, {:tuple, n, right_types}, context) do
     left_types
     |> Enum.zip(right_types)
-    |> Enum.any?(fn {left, right} -> subtype?(left, right, context) end)
+    |> Enum.all?(fn {left, right} -> subtype?(left, right, context) end)
   end
 
   def subtype?({:map, left_pairs}, {:map, right_pairs}, context) do
@@ -600,17 +648,15 @@ defmodule Module.Types.Unify do
   def to_union([type], _context), do: type
 
   def to_union(types, context) when types != [] do
-    flat_types = flatten_union(types)
-
-    case unique_super_types(flat_types, context) do
+    case unique_super_types(unnest_unions(types), context) do
       [type] -> type
       types -> {:union, types}
     end
   end
 
-  defp flatten_union(types) do
+  defp unnest_unions(types) do
     Enum.flat_map(types, fn
-      {:union, types} -> flatten_union(types)
+      {:union, types} -> unnest_unions(types)
       type -> [type]
     end)
   end
@@ -635,6 +681,204 @@ defmodule Module.Types.Unify do
     []
   end
 
+  ## Type lifting
+
+  @doc """
+  Lifts type variables to their inferred types from the context.
+  """
+  def lift_types(types, %{lifted_types: _} = context) do
+    Enum.map_reduce(types, context, &lift_type/2)
+  end
+
+  def lift_types(types, context) do
+    context = %{
+      types: context.types,
+      lifted_types: %{},
+      lifted_counter: 0
+    }
+
+    Enum.map_reduce(types, context, &lift_type/2)
+  end
+
+  # Lift type variable to its inferred (hopefully concrete) types from the context
+  defp lift_type({:var, var}, context) do
+    case context.lifted_types do
+      %{^var => lifted_var} ->
+        {{:var, lifted_var}, context}
+
+      %{} ->
+        case context.types do
+          %{^var => :unbound} ->
+            new_lifted_var(var, context)
+
+          %{^var => type} ->
+            if recursive_type?(type, [], context) do
+              new_lifted_var(var, context)
+            else
+              # Remove visited types to avoid infinite loops
+              # then restore after we are done recursing on vars
+              types = context.types
+              context = put_in(context.types[var], :unbound)
+              {type, context} = lift_type(type, context)
+              {type, %{context | types: types}}
+            end
+
+          %{} ->
+            new_lifted_var(var, context)
+        end
+    end
+  end
+
+  defp lift_type({:union, types}, context) do
+    {types, context} = Enum.map_reduce(types, context, &lift_type/2)
+    {{:union, types}, context}
+  end
+
+  defp lift_type({:tuple, n, types}, context) do
+    {types, context} = Enum.map_reduce(types, context, &lift_type/2)
+    {{:tuple, n, types}, context}
+  end
+
+  defp lift_type({:map, pairs}, context) do
+    {pairs, context} =
+      Enum.map_reduce(pairs, context, fn {kind, key, value}, context ->
+        {key, context} = lift_type(key, context)
+        {value, context} = lift_type(value, context)
+        {{kind, key, value}, context}
+      end)
+
+    {{:map, pairs}, context}
+  end
+
+  defp lift_type({:list, type}, context) do
+    {type, context} = lift_type(type, context)
+    {{:list, type}, context}
+  end
+
+  defp lift_type({:fun, clauses}, context) do
+    clauses =
+      Enum.map_reduce(clauses, context, fn {args, return}, context ->
+        {[return | args], context} = Enum.map_reduce([return | args], context, &lift_type/2)
+        {{args, return}, context}
+      end)
+
+    {{:fun, clauses}, context}
+  end
+
+  defp lift_type(other, context) do
+    {other, context}
+  end
+
+  defp new_lifted_var(original_var, context) do
+    types = Map.put(context.lifted_types, original_var, context.lifted_counter)
+    counter = context.lifted_counter + 1
+
+    type = {:var, context.lifted_counter}
+    context = %{context | lifted_types: types, lifted_counter: counter}
+    {type, context}
+  end
+
+  # TODO: Figure out function expansion
+
+  @doc """
+  Expand unions so that all unions are at the top level.
+
+      {integer() | float()} => {integer()} | {float()}
+  """
+  def flatten_union({:union, types}, context) do
+    Enum.flat_map(types, &flatten_union(&1, context))
+  end
+
+  def flatten_union(type, context) do
+    List.wrap(do_flatten_union(type, context))
+  end
+
+  def do_flatten_union({:tuple, num, types}, context) do
+    flatten_union_tuple(types, num, context, [])
+  end
+
+  def do_flatten_union({:list, type}, context) do
+    case do_flatten_union(type, context) do
+      {:union, union_types} -> Enum.map(union_types, &{:list, &1})
+      _type -> [{:list, type}]
+    end
+  end
+
+  def do_flatten_union({:map, pairs}, context) do
+    flatten_union_map(pairs, context, [])
+  end
+
+  def do_flatten_union({:var, var}, context) do
+    if looping_var?(var, context, []) do
+      {:var, var}
+    else
+      case context.types do
+        %{^var => :unbound} -> {:var, var}
+        %{^var => {:union, types}} -> Enum.map(types, &do_flatten_union(&1, context))
+        %{^var => type} -> do_flatten_union(type, context)
+      end
+    end
+  end
+
+  def do_flatten_union(type, _context) do
+    type
+  end
+
+  defp flatten_union_tuple([type | types], num, context, acc) do
+    case do_flatten_union(type, context) do
+      {:union, union_types} ->
+        Enum.flat_map(union_types, &flatten_union_tuple(types, num, context, [&1 | acc]))
+
+      type ->
+        flatten_union_tuple(types, num, context, [type | acc])
+    end
+  end
+
+  defp flatten_union_tuple([], num, _context, acc) do
+    [{:tuple, num, Enum.reverse(acc)}]
+  end
+
+  defp flatten_union_map([{kind, key, value} | pairs], context, acc) do
+    case do_flatten_union(key, context) do
+      {:union, union_types} ->
+        Enum.flat_map(union_types, &flatten_union_map_value(kind, &1, value, pairs, context, acc))
+
+      type ->
+        flatten_union_map_value(kind, type, value, pairs, context, acc)
+    end
+  end
+
+  defp flatten_union_map([], _context, acc) do
+    [{:map, Enum.reverse(acc)}]
+  end
+
+  defp flatten_union_map_value(kind, key, value, pairs, context, acc) do
+    case do_flatten_union(value, context) do
+      {:union, union_types} ->
+        Enum.flat_map(union_types, &flatten_union_map(pairs, context, [{kind, key, &1} | acc]))
+
+      value ->
+        flatten_union_map(pairs, context, [{kind, key, value} | acc])
+    end
+  end
+
+  defp looping_var?(var, context, parents) do
+    case context.types do
+      %{^var => :unbound} ->
+        false
+
+      %{^var => {:var, type}} ->
+        if var in parents do
+          true
+        else
+          looping_var?(type, context, [var | parents])
+        end
+
+      %{^var => _type} ->
+        false
+    end
+  end
+
   @doc """
   Formats types.
 
@@ -644,7 +888,7 @@ defmodule Module.Types.Unify do
   def format_type({:map, pairs}, true) do
     case List.keyfind(pairs, {:atom, :__struct__}, 1) do
       {:required, {:atom, :__struct__}, {:atom, struct}} ->
-        "%#{inspect(struct)}{}"
+        ["%", inspect(struct), "{}"]
 
       _ ->
         "map()"
@@ -652,24 +896,31 @@ defmodule Module.Types.Unify do
   end
 
   def format_type({:union, types}, simplify?) do
-    "#{Enum.map_join(types, " | ", &format_type(&1, simplify?))}"
+    types
+    |> Enum.map(&format_type(&1, simplify?))
+    |> Enum.intersperse(" | ")
   end
 
   def format_type({:tuple, _, types}, simplify?) do
-    "{#{Enum.map_join(types, ", ", &format_type(&1, simplify?))}}"
+    format =
+      types
+      |> Enum.map(&format_type(&1, simplify?))
+      |> Enum.intersperse(", ")
+
+    ["{", format, "}"]
   end
 
   def format_type({:list, type}, simplify?) do
-    "[#{format_type(type, simplify?)}]"
+    ["[", format_type(type, simplify?), "]"]
   end
 
   def format_type({:map, pairs}, false) do
     case List.keytake(pairs, {:atom, :__struct__}, 1) do
       {{:required, {:atom, :__struct__}, {:atom, struct}}, pairs} ->
-        "%#{inspect(struct)}{#{format_map_pairs(pairs)}}"
+        ["%", inspect(struct), "{", format_map_pairs(pairs), "}"]
 
       _ ->
-        "%{#{format_map_pairs(pairs)}}"
+        ["%{", format_map_pairs(pairs), "}"]
     end
   end
 
@@ -678,26 +929,84 @@ defmodule Module.Types.Unify do
   end
 
   def format_type({:var, index}, _simplify?) do
-    "var#{index}"
+    ["var", Integer.to_string(index + 1)]
+  end
+
+  def format_type({:fun, clauses}, simplify?) do
+    format =
+      Enum.map(clauses, fn {params, return} ->
+        params = Enum.intersperse(Enum.map(params, &format_type(&1, simplify?)), ", ")
+        params = if params == [], do: params, else: [params, " "]
+        return = format_type(return, simplify?)
+        [params, "-> ", return]
+      end)
+
+    ["(", Enum.intersperse(format, "; "), ")"]
   end
 
   def format_type(atom, _simplify?) when is_atom(atom) do
-    "#{atom}()"
+    [Atom.to_string(atom), "()"]
   end
 
   defp format_map_pairs(pairs) do
     {atoms, others} = Enum.split_with(pairs, &match?({:required, {:atom, _}, _}, &1))
     {required, optional} = Enum.split_with(others, &match?({:required, _, _}, &1))
 
-    Enum.map_join(atoms ++ required ++ optional, ", ", fn
+    (atoms ++ required ++ optional)
+    |> Enum.map(fn
       {:required, {:atom, atom}, right} ->
-        "#{atom}: #{format_type(right, false)}"
+        [Atom.to_string(atom), ": ", format_type(right, false)]
 
       {:required, left, right} ->
-        "#{format_type(left, false)} => #{format_type(right, false)}"
+        [format_type(left, false), " => ", format_type(right, false)]
 
       {:optional, left, right} ->
-        "optional(#{format_type(left, false)}) => #{format_type(right, false)}"
+        ["optional(", format_type(left, false), ") => ", format_type(right, false)]
     end)
+    |> Enum.intersperse(", ")
+  end
+
+  @doc """
+  Performs a depth-first, pre-order traversal of the type tree using an accumulator.
+  """
+  def walk({:map, pairs}, acc, fun) do
+    {pairs, acc} =
+      Enum.map_reduce(pairs, acc, fn {kind, key, value}, acc ->
+        {key, acc} = walk(key, acc, fun)
+        {value, acc} = walk(value, acc, fun)
+        {{kind, key, value}, acc}
+      end)
+
+    fun.({:map, pairs}, acc)
+  end
+
+  def walk({:union, types}, acc, fun) do
+    {types, acc} = Enum.map_reduce(types, acc, &walk(&1, &2, fun))
+    fun.({:union, types}, acc)
+  end
+
+  def walk({:tuple, num, types}, acc, fun) do
+    {types, acc} = Enum.map_reduce(types, acc, &walk(&1, &2, fun))
+    fun.({:tuple, num, types}, acc)
+  end
+
+  def walk({:list, type}, acc, fun) do
+    {type, acc} = walk(type, acc, fun)
+    fun.({:list, type}, acc)
+  end
+
+  def walk({:fun, clauses}, acc, fun) do
+    {clauses, acc} =
+      Enum.map_reduce(clauses, acc, fn {params, return}, acc ->
+        {params, acc} = Enum.map_reduce(params, acc, &walk(&1, &2, fun))
+        {return, acc} = walk(return, acc, fun)
+        {{params, return}, acc}
+      end)
+
+    fun.({:fun, clauses}, acc)
+  end
+
+  def walk(type, acc, fun) do
+    fun.(type, acc)
   end
 end
