@@ -17,10 +17,14 @@ defmodule Module.Types.TypesTest do
     end
   end
 
+  defmacro generated(ast) do
+    Macro.prewalk(ast, fn node -> Macro.update_meta(node, &([generated: true] ++ &1)) end)
+  end
+
   def __expr__({patterns, guards, body}) do
     with {:ok, _types, context} <-
            Pattern.of_head(patterns, guards, TypeHelper.new_stack(), TypeHelper.new_context()),
-         {:ok, _type, context} <- Expr.of_expr(body, TypeHelper.new_stack(), context) do
+         {:ok, _type, context} <- Expr.of_expr(body, :dynamic, TypeHelper.new_stack(), context) do
       case context.warnings do
         [warning] -> to_message(:warning, warning)
         _ -> :none
@@ -144,14 +148,14 @@ defmodule Module.Types.TypesTest do
       assert string == """
              incompatible types:
 
-                 {var0} !~ var0
+                 {var1} !~ var1
 
              in expression:
 
                  # types_test.ex:1
                  {var} = var
 
-             where "var" was given the type {var0} in:
+             where "var" was given the type {var1} in:
 
                  # types_test.ex:1
                  {var} = var
@@ -169,7 +173,7 @@ defmodule Module.Types.TypesTest do
              in expression:
 
                  # types_test.ex:1
-                 is_integer(var) and is_binary(var)
+                 is_binary(var)
 
              where "var" was given the type integer() in:
 
@@ -194,7 +198,7 @@ defmodule Module.Types.TypesTest do
              in expression:
 
                  # types_test.ex:1
-                 is_integer(x) and is_binary(y)
+                 is_binary(y)
 
              where "y" was given the same type as "x" in:
 
@@ -213,6 +217,32 @@ defmodule Module.Types.TypesTest do
              """
     end
 
+    test "warns on guards from cases unless generated" do
+      string =
+        warning(
+          [var],
+          [is_integer(var)],
+          case var do
+            _ when is_binary(var) -> :ok
+          end
+        )
+
+      assert is_binary(string)
+
+      string =
+        generated(
+          warning(
+            [var],
+            [is_integer(var)],
+            case var do
+              _ when is_binary(var) -> :ok
+            end
+          )
+        )
+
+      assert string == :none
+    end
+
     test "only show relevant traces in warning" do
       string = warning([x = y, z], [is_integer(x) and is_binary(y) and is_boolean(z)], {x, y, z})
 
@@ -224,7 +254,7 @@ defmodule Module.Types.TypesTest do
              in expression:
 
                  # types_test.ex:1
-                 is_integer(x) and is_binary(y) and is_boolean(z)
+                 is_binary(y)
 
              where "y" was given the same type as "x" in:
 
@@ -320,9 +350,116 @@ defmodule Module.Types.TypesTest do
                  <<foo::integer()>>
              """
     end
+
+    test "is_tuple warning" do
+      string = warning([foo], [is_tuple(foo)], {_} = foo)
+
+      assert string == """
+             incompatible types:
+
+                 tuple() !~ {dynamic()}
+
+             in expression:
+
+                 # types_test.ex:1
+                 {_} = foo
+
+             where "foo" was given the type tuple() in:
+
+                 # types_test.ex:1
+                 is_tuple(foo)
+
+             where "foo" was given the type {dynamic()} in:
+
+                 # types_test.ex:1
+                 {_} = foo
+
+             HINT: use pattern matching or "is_tuple(foo) and tuple_size(foo) == 1" to guard a sized tuple.
+             """
+    end
+
+    test "function call" do
+      string = warning([foo], [rem(foo, 2.0) == 0], foo)
+
+      assert string == """
+             expected Kernel.rem/2 to have signature:
+
+                 var1, float() -> dynamic()
+
+             but it has signature:
+
+                 integer(), integer() -> integer()
+
+             in expression:
+
+                 # types_test.ex:1
+                 rem(foo, 2.0)
+             """
+    end
+
+    test "operator call" do
+      string = warning([foo], [foo - :bar == 0], foo)
+
+      assert string == """
+             expected Kernel.-/2 to have signature:
+
+                 var1, :bar -> dynamic()
+
+             but it has signature:
+
+                 integer(), integer() -> integer()
+                 float(), integer() | float() -> float()
+                 integer() | float(), float() -> float()
+
+             in expression:
+
+                 # types_test.ex:1
+                 foo - :bar
+             """
+    end
   end
 
   describe "map warnings" do
+    test "handling of non-singleton types in maps" do
+      string =
+        warning(
+          [],
+          (
+            event = %{"type" => "order"}
+            %{"amount" => amount} = event
+            %{"user" => user} = event
+            %{"id" => user_id} = user
+            {:order, user_id, amount}
+          )
+        )
+
+      assert string == """
+             incompatible types:
+
+                 binary() !~ map()
+
+             in expression:
+
+                 # types_test.ex:5
+                 %{"id" => user_id} = user
+
+             where "user" was given the same type as "amount" in:
+
+                 # types_test.ex:4
+                 %{"user" => user} = event
+
+             where "user" was given the type map() in:
+
+                 # types_test.ex:5
+                 %{"id" => user_id} = user
+
+             where "amount" was given the type binary() in:
+
+                 # types_test.ex:3
+                 %{"amount" => amount} = event
+             """
+    end
+
     test "show map() when comparing against non-map" do
       string =
         warning(
@@ -421,7 +558,7 @@ defmodule Module.Types.TypesTest do
              """
     end
 
-    test "non-existant map field warning" do
+    test "non-existent map field warning" do
       string =
         warning(
           (
@@ -445,7 +582,7 @@ defmodule Module.Types.TypesTest do
              """
     end
 
-    test "non-existant struct field warning" do
+    test "non-existent struct field warning" do
       string =
         warning(
           [foo],
@@ -469,6 +606,71 @@ defmodule Module.Types.TypesTest do
                  %URI{} = foo
              """
     end
+
+    test "expands type variables" do
+      string =
+        warning(
+          [%{foo: key} = event, other_key],
+          [is_integer(key) and is_atom(other_key)],
+          %{foo: ^other_key} = event
+        )
+
+      assert string == """
+             incompatible types:
+
+                 %{foo: integer()} !~ %{foo: atom()}
+
+             in expression:
+
+                 # types_test.ex:3
+                 %{foo: ^other_key} = event
+
+             where "event" was given the type %{foo: integer(), optional(dynamic()) => dynamic()} in:
+
+                 # types_test.ex:1
+                 %{foo: key} = event
+
+             where "event" was given the type %{foo: atom(), optional(dynamic()) => dynamic()} in:
+
+                 # types_test.ex:3
+                 %{foo: ^other_key} = event
+             """
+    end
+
+    test "expands map when maps are nested" do
+      string =
+        warning(
+          [map1, map2],
+          (
+            [_var1, _var2] = [map1, map2]
+            %{} = map1
+            %{} = map2.subkey
+          )
+        )
+
+      assert string == """
+             incompatible types:
+
+                 %{subkey: var1, optional(dynamic()) => dynamic()} !~ %{optional(dynamic()) => dynamic()} | %{optional(dynamic()) => dynamic()}
+
+             in expression:
+
+                 # types_test.ex:5
+                 map2.subkey
+
+             where "map2" was given the type %{optional(dynamic()) => dynamic()} | %{optional(dynamic()) => dynamic()} in:
+
+                 # types_test.ex:3
+                 [_var1, _var2] = [map1, map2]
+
+             where "map2" was given the type %{subkey: var1, optional(dynamic()) => dynamic()} (due to calling var.field) in:
+
+                 # types_test.ex:5
+                 map2.subkey
+
+             HINT: "var.field" (without parentheses) implies "var" is a map() while "var.fun()" (with parentheses) implies "var" is an atom()
+             """
+    end
   end
 
   describe "regressions" do
@@ -485,5 +687,18 @@ defmodule Module.Types.TypesTest do
                end
              ) == :none
     end
+
+    test "no-recursion on guards with map fields" do
+      assert warning(
+               [assigns],
+               (
+                 variable_enum = assigns.variable_enum
+
+                 case true do
+                   _ when variable_enum != nil -> assigns.variable_enum
+                 end
+               )
+             ) == :none
+    end
   end
 end
