@@ -6,8 +6,8 @@ defmodule Mix.Tasks.Release do
   @moduledoc """
   Assembles a self-contained release for the current project:
 
-      MIX_ENV=prod mix release
-      MIX_ENV=prod mix release NAME
+      $ MIX_ENV=prod mix release
+      $ MIX_ENV=prod mix release NAME
 
   Once a release is assembled, it can be packaged and deployed to a
   target, as long as the target runs on the same operating system (OS)
@@ -34,7 +34,7 @@ defmodule Mix.Tasks.Release do
   and the value is a keyword list with the release configuration.
   Releasing a certain name is done with:
 
-      MIX_ENV=prod mix release demo
+      $ MIX_ENV=prod mix release demo
 
   If the given name does not exist, an error is raised.
 
@@ -87,8 +87,8 @@ defmodule Mix.Tasks.Release do
   Once a release is assembled, you can start it by calling
   `bin/RELEASE_NAME start` inside the release. In production, you would do:
 
-      MIX_ENV=prod mix release
-      _build/prod/rel/my_app/bin/my_app start
+      $ MIX_ENV=prod mix release
+      $ _build/prod/rel/my_app/bin/my_app start
 
   `bin/my_app start` will start the system connected to the current standard
   input/output, where logs are also written to by default. This is the
@@ -108,8 +108,8 @@ defmodule Mix.Tasks.Release do
   If you want to invoke specific modules and functions in your release,
   you can do so in two ways: using `eval` or `rpc`.
 
-      bin/RELEASE_NAME eval "IO.puts(:hello)"
-      bin/RELEASE_NAME rpc "IO.puts(:hello)"
+      $ bin/RELEASE_NAME eval "IO.puts(:hello)"
+      $ bin/RELEASE_NAME rpc "IO.puts(:hello)"
 
   The `eval` command starts its own instance of the VM but without
   starting any of the applications in the release and without starting
@@ -161,18 +161,18 @@ defmodule Mix.Tasks.Release do
 
   And to run them:
 
-      bin/RELEASE_NAME eval "MyApp.ReleaseTasks.eval_purge_stale_data()"
-      bin/RELEASE_NAME rpc "MyApp.ReleaseTasks.rpc_print_connected_users()"
+      $ bin/RELEASE_NAME eval "MyApp.ReleaseTasks.eval_purge_stale_data()"
+      $ bin/RELEASE_NAME rpc "MyApp.ReleaseTasks.rpc_print_connected_users()"
 
   ### Daemon mode (Unix-like)
 
   You can run the release in daemon mode with the command:
 
-      bin/RELEASE_NAME daemon_iex
+      $ bin/RELEASE_NAME daemon
 
   In daemon mode, the system is started on the background via
-  [run_erl](http://erlang.org/doc/man/run_erl.html). You may also
-  want to enable [heart](http://erlang.org/doc/man/heart.html)
+  [`run_erl`](https://erlang.org/doc/man/run_erl.html). You may also
+  want to enable [`heart`](https://erlang.org/doc/man/heart.html)
   in daemon mode so it automatically restarts the system in case
   of crashes. See the generated `releases/RELEASE_VSN/env.sh` file.
 
@@ -195,10 +195,10 @@ defmodule Mix.Tasks.Release do
 
   While daemons are not available on Windows, it is possible to install a
   released system as a service on Windows with the help of
-  [erlsrv](http://erlang.org/doc/man/erlsrv.html). This can be done by
+  [`erlsrv`](https://erlang.org/doc/man/erlsrv.html). This can be done by
   running:
 
-      bin/RELEASE_NAME install
+      $ bin/RELEASE_NAME install
 
   Once installed, the service must be explicitly managed via the `erlsrv`
   executable, which is included in the `erts-VSN/bin` directory.
@@ -207,8 +207,8 @@ defmodule Mix.Tasks.Release do
   For example, if you have a release named `demo`, you can install
   the service and then start it from the release root as follows:
 
-      bin/demo install
-      erts-VSN/bin/erlsrv.exs start demo_demo
+      $ bin/demo install
+      $ erts-VSN/bin/erlsrv.exe start demo_demo
 
   The name of the service is `demo_demo` because the name is built
   by concatenating the node name with the release name. Since Elixir
@@ -221,18 +221,20 @@ defmodule Mix.Tasks.Release do
 
   The following commands are supported by `bin/RELEASE_NAME`:
 
-      start        Starts the system
-      start_iex    Starts the system with IEx attached
-      daemon       Starts the system as a daemon (Unix-like only)
-      daemon_iex   Starts the system as a daemon with IEx attached (Unix-like only)
-      install      Installs this system as a Windows service (Windows only)
-      eval "EXPR"  Executes the given expression on a new, non-booted system
-      rpc "EXPR"   Executes the given expression remotely on the running system
-      remote       Connects to the running system via a remote shell
-      restart      Restarts the running system via a remote command
-      stop         Stops the running system via a remote command
-      pid          Prints the operating system PID of the running system via a remote command
-      version      Prints the release name and version to be booted
+  ```text
+  start        Starts the system
+  start_iex    Starts the system with IEx attached
+  daemon       Starts the system as a daemon (Unix-like only)
+  daemon_iex   Starts the system as a daemon with IEx attached (Unix-like only)
+  install      Installs this system as a Windows service (Windows only)
+  eval "EXPR"  Executes the given expression on a new, non-booted system
+  rpc "EXPR"   Executes the given expression remotely on the running system
+  remote       Connects to the running system via a remote shell
+  restart      Restarts the running system via a remote command
+  stop         Stops the running system via a remote command
+  pid          Prints the operating system PID of the running system via a remote command
+  version      Prints the release name and version to be booted
+  ```
 
   ## Deployments
 
@@ -248,7 +250,7 @@ defmodule Mix.Tasks.Release do
   the following must be the same between the host and the target:
 
     * Target architecture (for example, x86_64 or ARM)
-    * Target vendor + operating system  (for example, Windows, Linux, or Darwin/macOS)
+    * Target vendor + operating system (for example, Windows, Linux, or Darwin/macOS)
     * Target ABI (for example, musl or gnu)
 
   This is often represented in the form of target triples, for example,
@@ -304,22 +306,22 @@ defmodule Mix.Tasks.Release do
   compile the code and assemble the release on the target itself. It would
   be something like this:
 
-      git clone remote://path/to/my_app.git my_app_source
-      cd my_app_source
-      mix deps.get --only prod
-      MIX_ENV=prod mix release
-      _build/prod/rel/my_app/bin/my_app start
+      $ git clone remote://path/to/my_app.git my_app_source
+      $ cd my_app_source
+      $ mix deps.get --only prod
+      $ MIX_ENV=prod mix release
+      $ _build/prod/rel/my_app/bin/my_app start
 
   If you prefer, you can also compile the release to a separate directory,
   so you can erase all source after the release is assembled:
 
-      git clone remote://path/to/my_app.git my_app_source
-      cd my_app_source
-      mix deps.get --only prod
-      MIX_ENV=prod mix release --path ../my_app_release
-      cd ../my_app_release
-      rm -rf ../my_app_source
-      bin/my_app start
+      $ git clone remote://path/to/my_app.git my_app_source
+      $ cd my_app_source
+      $ mix deps.get --only prod
+      $ MIX_ENV=prod mix release --path ../my_app_release
+      $ cd ../my_app_release
+      $ rm -rf ../my_app_source
+      $ bin/my_app start
 
   However, this option can be expensive if you have multiple production
   nodes or if the release assembling process is a long one, as each node
@@ -364,10 +366,16 @@ defmodule Mix.Tasks.Release do
 
   The following options can be set inside your `mix.exs` on each release definition:
 
-    * `:applications` - a keyword list that configures and adds new applications
-      to the release. The key is the application name and the value is one of:
+    * `:applications` - a keyword list with application names as keys and their
+      mode as value. By default `:applications` includes the current application and
+      all applications the current application depends on, recursively. You can include
+      new applications or change the mode of existing ones by listing them here.
 
-        * `:permanent` - the application is started and the node shuts down
+      The order of the applications given will be preserved as much as possible, with
+      only `:kernel`, `:stdlib`, `:sasl`, and `:elixir` listed before the given application
+      list. The supported values are:
+
+        * `:permanent` (default) - the application is started and the node shuts down
           if the application terminates, regardless of reason
         * `:transient` - the application is started and the node shuts down
           if the application terminates abnormally
@@ -377,26 +385,13 @@ defmodule Mix.Tasks.Release do
         * `:none` - the application is part of the release but it is neither
           loaded nor started
 
-      All applications default to `:permanent`.
-
-      By default `:applications` includes the current application and all
-      applications the current application depends on, recursively.
-      You can include new applications or change the mode of
-      existing ones by listing them here. The order of the applications given
-      in `:applications` will be preserved as much as possible, with only
-      `:kernel`, `:stdlib`, `:sasl`, and `:elixir` listed before the given
-      application list.
-
-      Releases assembled from an umbrella project require this configuration
-      to be explicitly given.
-
     * `:strip_beams` - controls if BEAM files should have their debug information,
       documentation chunks, and other non-essential metadata removed. Defaults to
-      `true`. Maybe be set to `false` to disable striping. Also accepts
+      `true`. May be set to `false` to disable stripping. Also accepts
       `[keep: ["Docs", "Dbgi"]]` to keep certain chunks that are usually stripped.
 
     * `:cookie` - a string representing the Erlang Distribution cookie. If this
-      option is not set, a random cookie is  written to the `releases/COOKIE` file
+      option is not set, a random cookie is written to the `releases/COOKIE` file
       when the first release is assembled. At runtime, we will first attempt
       to fetch the cookie from the `RELEASE_COOKIE` environment variable and
       then we'll read the `releases/COOKIE` file.
@@ -449,8 +444,8 @@ defmodule Mix.Tasks.Release do
           ]
 
     * `:rel_templates_path` - the path to find template files that are copied to
-      the release, such as "vm.args.eex", "env.sh.eex" (or "env.bat.eex"), and
-      "overlays". Defaults to "rel" in the project root.
+      the release, such as "vm.args.eex", "remote.vm.args.eex", "env.sh.eex"
+      (or "env.bat.eex"), and "overlays". Defaults to "rel" in the project root.
 
     * `:overlays` - a list of directories with extra files to be copied
       as is to the release. The "overlays" directory at `:rel_templates_path`
@@ -460,6 +455,14 @@ defmodule Mix.Tasks.Release do
     * `:steps` - a list of steps to execute when assembling the release. See
       the "Steps" section for more information.
 
+    * `:skip_mode_validation_for` - a list of application names
+      (atoms) specifying applications to skip strict validation of
+      "unsafe" modes. An "unsafe" case is when a parent application
+      mode is `:permanent` but one of the applications it depends on
+      is set to `:load`. Use this with care, as a release with
+      invalid modes may no longer boot without additional tweaks.
+      Defaults to `[]`.
+
   Note each release definition can be given as an anonymous function. This
   is useful if some release attributes are expensive to compute:
 
@@ -520,20 +523,16 @@ defmodule Mix.Tasks.Release do
   ### vm.args and env.sh (env.bat)
 
   Developers may want to customize the VM flags and environment variables
-  given when the release starts. This is typically done by customizing
-  two files inside your release: `releases/RELEASE_VSN/vm.args` and
-  `releases/RELEASE_VSN/env.sh` (or `env.bat` on Windows).
+  given when the release starts. The simplest way to customize those files
+  is by running `mix release.init`. The Mix task will copy custom
+  `rel/vm.args.eex`, `rel/remote.vm.args.eex`,  `rel/env.sh.eex`, and
+  `rel/env.bat.eex` files to your project root. You can modify those files
+  and they will be evaluated every time you perform a new release. Those
+  files are regular EEx templates and they have a single assign, called
+  `@release`, with the `Mix.Release` struct.
 
-  However, instead of modifying those files after the release is built,
-  the simplest way to customize those files is by running `mix release.init`.
-  The Mix task will copy custom `rel/vm.args.eex`, `rel/env.sh.eex`, and
-  `rel/env.bat.eex` files to your project root. You can modify those
-  files and they will be evaluated every time you perform a new release.
-  Those files are regular EEx templates and they have a single assign,
-  called `@release`, with the `Mix.Release` struct.
-
-  The `vm.args` file may contain any of the VM flags accepted by the [`erl`
-  command](http://erlang.org/doc/man/erl.html).
+  The `vm.args` and `remote.vm.args` files may contain any of the VM flags
+  accepted by the [`erl` command](https://erlang.org/doc/man/erl.html).
 
   The `env.sh` and `env.bat` is used to set environment variables.
   In there, you can set vars such as `RELEASE_NODE`, `RELEASE_COOKIE`,
@@ -543,19 +542,21 @@ defmodule Mix.Tasks.Release do
   `RELEASE_COMMAND` have already been set, so you can rely on them.
   See the section on environment variables for more information.
 
-  Furthermore, while `vm.args` is static, you can use `env.sh` and
-  `env.bat` to dynamically set VM options. For example, if you want
-  to make sure the Erlang Distribution listens only on a given port
-  known at runtime, you can set the following:
+  Furthermore, while the `vm.args` files are static, you can use
+  `env.sh` and `env.bat` to dynamically set VM options. For example,
+  if you want to make sure the Erlang Distribution listens only on
+  a given port known at runtime, you can set the following:
 
-      case $RELEASE_COMMAND in
-        start*|daemon*)
-          ELIXIR_ERL_OPTIONS="-kernel inet_dist_listen_min $BEAM_PORT inet_dist_listen_max $BEAM_PORT"
-          export ELIXIR_ERL_OPTIONS
-          ;;
-        *)
-          ;;
-      esac
+  ```bash
+  case $RELEASE_COMMAND in
+    start*|daemon*)
+      ELIXIR_ERL_OPTIONS="-kernel inet_dist_listen_min $BEAM_PORT inet_dist_listen_max $BEAM_PORT"
+      export ELIXIR_ERL_OPTIONS
+      ;;
+    *)
+      ;;
+  esac
+  ```
 
   Note we only set the port on start/daemon commands. If you also limit
   the port on other commands, such as `rpc`, then you will be unable
@@ -564,36 +565,37 @@ defmodule Mix.Tasks.Release do
 
   On Windows, your `env.bat` would look like this:
 
-      IF NOT %RELEASE_COMMAND:start=%==%RELEASE_COMMAND% (
-        set ELIXIR_ERL_OPTIONS="-kernel inet_dist_listen_min %BEAM_PORT% inet_dist_listen_max %BEAM_PORT%"
-      )
+  ```bash
+  IF NOT %RELEASE_COMMAND:start=%==%RELEASE_COMMAND% (
+    set ELIXIR_ERL_OPTIONS="-kernel inet_dist_listen_min %BEAM_PORT% inet_dist_listen_max %BEAM_PORT%"
+  )
+  ```
 
   ## Application configuration
 
-  Releases provides two mechanisms for configuring OTP applications:
-  build-time and runtime.
+  Mix provides two mechanisms for configuring the application environment
+  of your application and your dependencies: build-time and runtime. On this
+  section, we will learn how those mechanisms apply to releases. An introduction
+  to this topic can be found in the "Configuration" section of the `Mix` module.
 
   ### Build-time configuration
 
-  Whenever you invoke a `mix` command, Mix loads the configuration
-  in `config/config.exs`, if said file exists. It is common for the
-  `config/config.exs` file itself to import other configuration based
-  on the current `MIX_ENV`, such as `config/dev.exs`, `config/test.exs`,
-  and `config/prod.exs`. We say that this configuration is a build-time
-  configuration as it is evaluated whenever you compile your code or
-  whenever you assemble the release.
+  Whenever you invoke a `mix` command, Mix loads the configuration in
+  `config/config.exs`, if said file exists. We say that this configuration
+  is a build-time configuration as it is evaluated whenever you compile your
+  code or whenever you assemble the release.
 
   In other words, if your configuration does something like:
 
+      import Config
       config :my_app, :secret_key, System.fetch_env!("MY_APP_SECRET_KEY")
 
   The `:secret_key` key under `:my_app` will be computed on the
-  host machine, whenever the release is built. Setting the
-  `MY_APP_SECRET_KEY` right before starting your release will have
-  no effect.
-
-  Luckily, releases also provide runtime configuration, which we will
-  see next.
+  host machine, whenever the release is built. Therefore if the machine
+  assembling the release not have access to all environment variables used
+  to run your code, loading the configuration will fail as the environment
+  variable is missing. Luckily, Mix also provides runtime configuration,
+  which should be preferred and we will see next.
 
   ### Runtime configuration
 
@@ -642,10 +644,10 @@ defmodule Mix.Tasks.Release do
 
   Releases also supports custom mechanisms, called config providers, to load
   any sort of runtime configuration to the system while it boots. For instance,
-  if you need to access a vault or load configuration from a JSON file, it
-  can be achieved with config providers. The runtime configuration outlined
-  in the previous section, which is handled by the `Config.Reader` provider.
-  See the `Config.Provider` module for more information and more examples.
+  if you need to access a vault or load configuration from a JSON file, it can
+  be achieved with config providers. The runtime configuration outlined in the
+  previous section is handled by the `Config.Reader` provider. See the
+  `Config.Provider` module for more information and more examples.
 
   The following options can be set inside your releases key in your `mix.exs`
   to control how config providers work:
@@ -671,7 +673,7 @@ defmodule Mix.Tasks.Release do
     * `:start_distribution_during_config` - if `:reboot_system_after_config` is
       set, releases only start the Erlang VM distribution features after the config
       files are evaluated. You can set it to `true` if you need distribution during
-      configuration. Defaults to `false` from Erlang/OTP 22+.
+      configuration. Defaults to `false`.
 
     * `:config_providers` - a list of tuples with custom config providers.
       See `Config.Provider` for more information. Defaults to `[]`.
@@ -691,9 +693,11 @@ defmodule Mix.Tasks.Release do
       detect you are inside a release, you can check for release specific
       environment variables, such as `RELEASE_NODE` or `RELEASE_MODE`
 
-    * `rel/vm.args.eex` - a template file that is copied into every release
-      and provides static configuration of the Erlang Virtual Machine and
-      other runtime flags
+    * `rel/vm.args.eex` and `rel/remote.vm.args.eex` - template files that
+      are copied into every release and provides static configuration of the
+      Erlang Virtual Machine and other runtime flags. `vm.args` runs on
+      `start`, `daemon`, and `eval` commands. `remote.vm.args` configures
+      the VM for `remote` and `rpc` commands
 
     * `rel/env.sh.eex` and `rel/env.bat.eex` - template files that are copied
       into every release and are executed on every command to set up environment
@@ -703,33 +707,42 @@ defmodule Mix.Tasks.Release do
 
   A release is organized as follows:
 
-      bin/
-        RELEASE_NAME
-      erts-ERTS_VSN/
-      lib/
-        APP_NAME-APP_VSN/
-          ebin/
-          include/
-          priv/
-      releases/
-        RELEASE_VSN/
-          consolidated/
-          elixir
-          elixir.bat
-          env.bat
-          env.sh
-          iex
-          iex.bat
-          runtime.exs
-          start.boot
-          start.script
-          start_clean.boot
-          start_clean.script
-          sys.config
-          vm.args
-        COOKIE
-        start_erl.data
-      tmp/
+  ```text
+  bin/
+    RELEASE_NAME
+  erts-ERTS_VSN/
+  lib/
+    APP_NAME-APP_VSN/
+      ebin/
+      include/
+      priv/
+  releases/
+    RELEASE_VSN/
+      consolidated/
+      elixir
+      elixir.bat
+      env.bat
+      env.sh
+      iex
+      iex.bat
+      remote.vm.args
+      runtime.exs
+      start.boot
+      start.script
+      start_clean.boot
+      start_clean.script
+      sys.config
+      vm.args
+    COOKIE
+    start_erl.data
+  tmp/
+  ```
+
+  We document this structure for completeness. In practice, developers
+  should not modify any of those files after the release is assembled.
+  Instead use env scripts, custom config provider, overlays, and all
+  other mechanisms described in this guide to configure how your release
+  works.
 
   ## Environment variables
 
@@ -756,15 +769,18 @@ defmodule Mix.Tasks.Release do
       release. The custom value must be an existing release version in
       the `releases/` directory
 
+    * `RELEASE_PROG` - the command line executable used to start the release
+
   The following variables can be set before you invoke the release or
   inside `env.sh` and `env.bat`:
 
     * `RELEASE_COOKIE` - the release cookie. By default uses the value
       in `releases/COOKIE`. It can be set to a custom value
 
-    * `RELEASE_NODE` - the release node name, in the format `name@host`.
-      It can be set to a custom value. The name part must be made only
-      of letters, digits, underscores, and hyphens
+    * `RELEASE_NODE` - the release node name, in the format `name` or
+      optionally `name@host` if running in distributed mode. It can be
+      set to a custom value. The name part must be made only of letters,
+      digits, underscores, and hyphens
 
     * `RELEASE_SYS_CONFIG` - the location of the sys.config file. It can
       be set to a custom path and it must not include the `.config` extension
@@ -772,6 +788,9 @@ defmodule Mix.Tasks.Release do
     * `RELEASE_VM_ARGS` - the location of the vm.args file. It can be set
       to a custom path
 
+    * `RELEASE_REMOTE_VM_ARGS` - the location of the remote.vm.args file.
+      It can be set to a custom path
+
     * `RELEASE_TMP` - the directory in the release to write temporary
       files to. It can be set to a custom directory. It defaults to
       `$RELEASE_ROOT/tmp`
@@ -784,9 +803,7 @@ defmodule Mix.Tasks.Release do
       May be `name` (long names), `sname` (short names) or `none`
       (distribution is not started automatically). Defaults to
       `sname` which allows access only within the current system.
-      `name` allows external connections. If `name` is used and you are
-      not running on Erlang/OTP 22 or later, you must set `RELEASE_NODE`
-      to `RELEASE_NAME@127.0.0.1` with an IP or a known host
+      `name` allows external connections
 
     * `RELEASE_BOOT_SCRIPT` - the name of the boot script to use when starting
       the release. This script is used when running commands such as `start` and
@@ -809,11 +826,13 @@ defmodule Mix.Tasks.Release do
   your release and the starting point for each release. For example,
   imagine this umbrella applications:
 
-      my_app_umbrella/
-        apps/
-          my_app_core/
-          my_app_event_processing/
-          my_app_web/
+  ```text
+  my_app_umbrella/
+    apps/
+      my_app_core/
+      my_app_event_processing/
+      my_app_web/
+  ```
 
   where both `my_app_event_processing` and `my_app_web` depend on
   `my_app_core` but they do not depend on each other.
@@ -924,7 +943,7 @@ defmodule Mix.Tasks.Release do
   If you were to perform a hot code upgrade in such an application, it would
   crash, because in the initial version the state was just a counter
   but in the new version the state is a tuple. Furthermore, you changed
-  the format of the `call` message from `:bump` to  `{:bump, by}` and
+  the format of the `call` message from `:bump` to `{:bump, by}` and
   the process may have both old and new messages temporarily mixed, so
   we need to handle both. The final version would be:
 
@@ -962,13 +981,13 @@ defmodule Mix.Tasks.Release do
   hot code upgrade it. This is one of the many steps necessary
   to perform hot code upgrades and it must be taken into account by
   every process and application being upgraded in the system.
-  The [`.appup` cookbook](http://erlang.org/doc/design_principles/appup_cookbook.html)
+  The [`.appup` cookbook](https://erlang.org/doc/design_principles/appup_cookbook.html)
   provides a good reference and more examples.
 
   Once `.appup`s are created, the next step is to create a `.relup`
   file with all instructions necessary to update the release itself.
   Erlang documentation does provide a chapter on
-  [Creating and Upgrading a Target System](http://erlang.org/doc/system_principles/create_target.html).
+  [Creating and upgrading a target system](https://erlang.org/doc/system_principles/create_target.html).
   [Learn You Some Erlang has a chapter on hot code upgrades](https://learnyousomeerlang.com/relups).
 
   Overall, there are many steps, complexities and assumptions made
@@ -1168,12 +1187,14 @@ defmodule Mix.Tasks.Release do
       end
 
     vm_args_path = Path.join(version_path, "vm.args")
+    remote_vm_args_path = Path.join(version_path, "remote.vm.args")
     cookie_path = Path.join(release.path, "releases/COOKIE")
     start_erl_path = Path.join(release.path, "releases/start_erl.data")
     config_provider_path = {:system, "RELEASE_SYS_CONFIG", ".config"}
 
     with :ok <- make_boot_scripts(release, version_path, consolidation_path),
          :ok <- make_vm_args(release, vm_args_path),
+         :ok <- make_vm_args(release, remote_vm_args_path),
          :ok <- Mix.Release.make_sys_config(release, sys_config, config_provider_path),
          :ok <- Mix.Release.make_cookie(release, cookie_path),
          :ok <- Mix.Release.make_start_erl(release, start_erl_path) do
@@ -1205,7 +1226,7 @@ defmodule Mix.Tasks.Release do
           {default_path, false}
 
         File.exists?(deprecated_path) ->
-          # TODO: Warn from Elixir v1.13 onwards
+          # TODO: Warn from Elixir v1.14 onwards
           {deprecated_path, true}
 
         true ->
@@ -1259,7 +1280,7 @@ defmodule Mix.Tasks.Release do
   end
 
   defp make_vm_args(release, path) do
-    vm_args_template = Mix.Release.rel_templates_path(release, "vm.args.eex")
+    vm_args_template = Mix.Release.rel_templates_path(release, "#{Path.basename(path)}.eex")
 
     if File.exists?(vm_args_template) do
       copy_template(vm_args_template, path, [release: release], force: true)
@@ -1415,15 +1436,17 @@ defmodule Mix.Tasks.Release do
     ]
   end
 
+  @erts_bin [~s[ERTS_BIN="$ERTS_BIN"], ~s[ERTS_BIN=!ERTS_BIN!]]
+
   defp replace_erts_bin(contents, release, new_path) do
     if release.erts_source do
-      String.replace(contents, ~s[ERTS_BIN=], ~s[ERTS_BIN=#{new_path}])
+      String.replace(contents, @erts_bin, ~s[ERTS_BIN=#{new_path}])
     else
       contents
     end
   end
 
-  defp executable!(path), do: File.chmod!(path, 0o744)
+  defp executable!(path), do: File.chmod!(path, 0o755)
 
   # Helper functions
 
